#include <node.h>
#include <stdlib.h>
#include <malloc.h>
#include <cstring>
#include <sstream>
#include <ostream>
#include "NodeFastbit.h"
#include "import/fastbit-2.0.3/src/ibis.h"
#include "import/fastbit-2.0.3/src/query.h"
#include "import/fastbit-2.0.3/src/bundle.h"

extern "C" {
    /// The object underlying the FastBit query handle.
    struct FastBitQuery {
        const ibis::part *t; ///!< The ibis::part this query refers to.
        ibis::query q; ///!< The ibis::query object
        typedef std::map< int, void* > typeValues;
        typedef std::map< const char*, typeValues*, ibis::lessi > valList;
        /// List of values that has been selected and sent to user.
        valList vlist;

        /// For storing null-terminated strings.
        struct NullTerminatedStrings {
            const char **pointers; ///!< The pointer passed to the caller.
            std::vector<std::string> *values; ///!< Actual string values.
        }; // NullTerminatedStrings
    };

    /// A @c FastBitResultSet holds the results of a query in memory and
    /// provides a row-oriented access mechanism for the results.
    ///@note An important limitation is that the current implementation
    /// requires all selected values to be in memory.
    struct FastBitResultSet {
        /// The ibis::query::result object to hold the results in memory.
        ibis::query::result *results;
        /// A place-holder for all the string objects.
        std::vector<std::string> strbuf;
    };
}
using namespace v8;

Persistent<Function> NodeFastbit::constructor;

NodeFastbit::NodeFastbit() {
};

NodeFastbit::~NodeFastbit() {
};

void NodeFastbit::Init( Handle<Object> exports ) {
    Isolate* isolate = Isolate::GetCurrent();

    // prep constructor template
    Local<FunctionTemplate> tpl = FunctionTemplate::New( isolate, New );
    tpl->SetClassName( String::NewFromUtf8(isolate, "NodeFastbit" ) );
    tpl->InstanceTemplate()->SetInternalFieldCount( 1 );

    // api methods
    //
    NODE_SET_PROTOTYPE_METHOD( tpl, "build_indexes", build_indexes );
    NODE_SET_PROTOTYPE_METHOD( tpl, "purge_indexes", purge_indexes );
    NODE_SET_PROTOTYPE_METHOD( tpl, "build_index", build_index );
    NODE_SET_PROTOTYPE_METHOD( tpl, "purge_index", purge_index );
    NODE_SET_PROTOTYPE_METHOD( tpl, "reorder_partition", reorder_partition );

    NODE_SET_PROTOTYPE_METHOD( tpl, "build_query", build_query );
    NODE_SET_PROTOTYPE_METHOD( tpl, "destroy_query", destroy_query );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_result_columns", get_result_columns );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_result_rows", get_result_rows );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_result_row_ids", get_result_row_ids );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_select_clause", get_select_clause );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_from_clause", get_from_clause );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_where_clause", get_where_clause );

    NODE_SET_PROTOTYPE_METHOD( tpl, "get_qualified_floats", get_qualified_floats );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_qualified_doubles", get_qualified_doubles );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_qualified_bytes", get_qualified_bytes );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_qualified_shorts", get_qualified_shorts );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_qualified_ints", get_qualified_ints );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_qualified_longs", get_qualified_longs );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_qualified_ubytes", get_qualified_ubytes );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_qualified_ushorts", get_qualified_ushorts );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_qualified_uints", get_qualified_uints );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_qualified_ulongs", get_qualified_ulongs );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_qualified_strings", get_qualified_strings );

    NODE_SET_PROTOTYPE_METHOD( tpl, "build_result_set", build_result_set );
    NODE_SET_PROTOTYPE_METHOD( tpl, "destroy_result_set", destroy_result_set );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_next", result_set_next );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_next_bundle", result_set_next_bundle );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_get_int", result_set_get_int );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_get_unsigned", result_set_get_unsigned );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_get_long", result_set_get_long );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_get_float", result_set_get_float );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_get_double", result_set_get_double );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_get_string", result_set_get_string );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_getInt", result_set_getInt );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_getUnsigned", result_set_getUnsigned );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_getLong", result_set_getLong );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_getFloat", result_set_getFloat );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_getDouble", result_set_getDouble );
    NODE_SET_PROTOTYPE_METHOD( tpl, "result_set_getString", result_set_getString );
    NODE_SET_PROTOTYPE_METHOD( tpl, "flush_buffer", flush_buffer );
    NODE_SET_PROTOTYPE_METHOD( tpl, "add_values", add_values );
    NODE_SET_PROTOTYPE_METHOD( tpl, "rows_in_partition", rows_in_partition );
    NODE_SET_PROTOTYPE_METHOD( tpl, "columns_in_partition", columns_in_partition );

    NODE_SET_PROTOTYPE_METHOD( tpl, "init", init );
    NODE_SET_PROTOTYPE_METHOD( tpl, "cleanup", cleanup );
    NODE_SET_PROTOTYPE_METHOD( tpl, "set_verbose_level", set_verbose_level );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_verbose_level", get_verbose_level );
    NODE_SET_PROTOTYPE_METHOD( tpl, "set_logfile", set_logfile );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_logfile", get_logfile );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_logfilepointer", get_logfilepointer );
    NODE_SET_PROTOTYPE_METHOD( tpl, "read_clock", read_clock );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_version_string", get_version_string );
    NODE_SET_PROTOTYPE_METHOD( tpl, "get_version_number", get_version_number );

    constructor.Reset(isolate, tpl->GetFunction());
    exports->Set( String::NewFromUtf8( isolate, "NodeFastbit" ), tpl->GetFunction() );
};

void NodeFastbit::New( const FunctionCallbackInfo<Value> &args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope(isolate);

    if (args.IsConstructCall()) {
        double value = args[0]->IsUndefined() ? 0 : args[0]->NumberValue();
        NodeFastbit* obj = new NodeFastbit();
        obj->Wrap(args.This());
        args.GetReturnValue().Set(args.This());
    } else {
        const int argc = 1;
        Local<Value> argv[argc] = { args[0] };
        Local<Function> cons = Local<Function>::New(isolate, constructor);
        args.GetReturnValue().Set(cons->NewInstance(argc, argv));
    }
}

void NodeFastbit::build_indexes( const FunctionCallbackInfo<Value> &args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }
    const char* dirName = getChar( args[0] );
    const char* optName = getChar( args[1] );
    int result = fastbit_build_indexes( dirName, optName );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::purge_indexes( const FunctionCallbackInfo<Value> &args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }
    const char* dirName = getChar( args[0] );
    int result = fastbit_purge_indexes( dirName );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::build_index( const FunctionCallbackInfo<Value> &args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 3 ) ) {
      return;
    }

    const char *indexLocation = getChar( args[0] );
    const char *cname = getChar( args[1] );
    const char *indexOptions = getChar( args[2]);

    int result = fastbit_build_index( indexLocation, cname, indexOptions );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::purge_index( const FunctionCallbackInfo<Value> &args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    const char *indexLocation = getChar( args[0] );
    const char *cname = getChar( args[1] );

    int result = fastbit_purge_index( indexLocation, cname );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::reorder_partition( const FunctionCallbackInfo<Value> &args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    const char *dir = getChar( args[0] );

    int result = fastbit_reorder_partition( dir );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::build_query( const FunctionCallbackInfo<Value> &args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 4 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );
    const char *selectClause = getChar( args[1] );
    const char *indexLocation = getChar( args[2] );
    const char *queryConditions = getChar( args[3] );

    std::pair< std::string, FastBitQuery* > builtQuery( queryHandle,
        fastbit_build_query( selectClause, indexLocation, queryConditions ));
    obj->m_queryHandles.insert( builtQuery );
    args.GetReturnValue().Set( args[0]->ToString() );
}

void NodeFastbit::destroy_query( const FunctionCallbackInfo<Value> &args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator it;
    it = obj->m_queryHandles.find( queryHandle );
    if (it != obj->m_queryHandles.end()) {
        FastBitQuery* foundQuery = it->second;
        int result = fastbit_destroy_query( foundQuery );
        obj->m_queryHandles.erase( queryHandle );
        args.GetReturnValue().Set( Integer::New( isolate, result ) );
    }
}

void NodeFastbit::get_result_columns( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator it;
    it = obj->m_queryHandles.find( queryHandle );
    if (it != obj->m_queryHandles.end()) {
        int result = fastbit_get_result_columns( it->second );
        args.GetReturnValue().Set( Integer::New( isolate, result ) );
    }
}

void NodeFastbit::get_result_rows( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator it;
    it = obj->m_queryHandles.find( queryHandle );
    if (it != obj->m_queryHandles.end()) {
        int result = fastbit_get_result_rows( it->second );
        args.GetReturnValue().Set( Integer::New( isolate, result ) );
    }
}

void NodeFastbit::get_result_row_ids( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator it;
    it = obj->m_queryHandles.find( queryHandle );
    if (it != obj->m_queryHandles.end()) {
        if (!it->second)
          return;
        int result = -1;
        // TRICKY: I rather not call two functions to accomplish this,
        //   but the return data is a bit different for the node.js side,
        //   as it is not returning a pointer, but an array.
        int resSize = fastbit_get_result_rows( it->second ) ;
        if ( resSize == 0 )
          return;

        Local<Object> returnObj = Object::New( isolate );
        uint32_t *ids = new uint32_t[sizeof(uint32_t)*resSize];
        result = fastbit_get_result_row_ids( it->second, ids );

        //TODO: make sure the return object has property "ret"
        returnObj->Set( String::NewFromUtf8( isolate, "ret" ),
                        Integer::New( isolate, result) );

        std::vector<uint32_t> res;

        for ( int i = 0; i < resSize; i++ ) {
            res.push_back( *ids );
            ids++;
        }
        ids-=resSize;
        delete[] ids;

        Local<Array> array = Array::New( isolate, resSize );

        uint32_t idx = 0;
        for (std::vector<uint32_t>::iterator it = res.begin(); it != res.end(); ++it) {
            array->Set( idx, Number::New(isolate, *it) );
            idx++;
        }
        returnObj->Set( String::NewFromUtf8( isolate, "ids" ), array );
        args.GetReturnValue().Set( returnObj );
    }
}

void NodeFastbit::get_select_clause( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator it;
    it = obj->m_queryHandles.find( queryHandle );
    if (it != obj->m_queryHandles.end()) {
        const char* qresult = fastbit_get_select_clause( it->second );
        std::string resStr( qresult );
        args.GetReturnValue().Set( String::NewFromUtf8( isolate, resStr.c_str() ) );
    }
}

void NodeFastbit::get_from_clause( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator it;
    it = obj->m_queryHandles.find( queryHandle );
    if (it != obj->m_queryHandles.end()) {
        const char* qresult = fastbit_get_from_clause( it->second );
        std::string resStr( qresult );
        args.GetReturnValue().Set( String::NewFromUtf8( isolate, resStr.c_str() ) );
    }
}

void NodeFastbit::get_where_clause( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator it;
    it = obj->m_queryHandles.find( queryHandle );
    if (it != obj->m_queryHandles.end()) {
        const char* qresult = fastbit_get_where_clause( it->second );
        std::string resStr( qresult );
        args.GetReturnValue().Set( String::NewFromUtf8( isolate, resStr.c_str() ) );
    }
}

void NodeFastbit::get_qualified_floats( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator q;
    q = obj->m_queryHandles.find( queryHandle );
    if (q == obj->m_queryHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );

    const float *qresult =
        fastbit_get_qualified_floats( q->second, cname );

    if ( qresult == 0 )
      return;

    std::vector<float> res;
    // TRICKY: I rather not call two functions to accomplish this,
    //   but the return data is a bit different for the node.js side,
    //   as it is not returning a pointer, but an array.
    int resSize = fastbit_get_result_rows( q->second ) ;

    if ( resSize == 0 )
      return;

    for ( int i = 0; i < resSize; i++ ) {
        res.push_back( *qresult );
        qresult++;
    }

    Local<Array> array = Array::New( isolate, resSize );

    uint32_t idx = 0;
    for (std::vector<float>::iterator it = res.begin(); it != res.end(); ++it) {
        array->Set( idx, Number::New(isolate, *it) );
        idx++;
    }
    args.GetReturnValue().Set( array );
}

void NodeFastbit::get_qualified_doubles( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator q;
    q = obj->m_queryHandles.find( queryHandle );
    if (q == obj->m_queryHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );
    const double *qresult =
        fastbit_get_qualified_doubles( q->second, cname );

    if ( qresult == 0 )
      return;

    std::vector<double> res;
    int resSize = fastbit_get_result_rows( q->second ) ;

    if ( resSize == 0 )
      return;

    for ( int i = 0; i < resSize; i++ ) {
        res.push_back( *qresult );
        qresult++;
    }

    Local<Array> array = Array::New( isolate, resSize );

    uint32_t idx = 0;
    for (std::vector<double>::iterator it = res.begin(); it != res.end(); ++it) {
        array->Set( idx, Number::New(isolate, *it) );
        idx++;
    }
    args.GetReturnValue().Set( array );
}

void NodeFastbit::get_qualified_bytes( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator q;
    q = obj->m_queryHandles.find( queryHandle );
    if (q == obj->m_queryHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );
    const signed char* qresult =
        fastbit_get_qualified_bytes( q->second, cname );

    if ( qresult == 0 )
      return;

    std::vector<const signed char*> res;
    int resSize = fastbit_get_result_rows( q->second ) ;

    if ( resSize == 0 )
      return;

    for ( int i = 0; i < resSize; i++ ) {
        res.push_back( qresult );
        qresult++;
    }

    Local<Array> array = Array::New( isolate, resSize );

    uint32_t idx = 0;
    for (std::vector<const signed char*>::iterator it = res.begin();
         it != res.end(); ++it) {
        std::ostringstream ss;
        ss << *it;
        array->Set( idx, String::NewFromUtf8( isolate, ss.str().c_str() ) );
        idx++;
    }
    args.GetReturnValue().Set( array );
}

void NodeFastbit::get_qualified_shorts( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator q;
    q = obj->m_queryHandles.find( queryHandle );
    if (q == obj->m_queryHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );
    const int16_t *qresult =
        fastbit_get_qualified_shorts( q->second, cname );

    if ( qresult == 0 )
      return;

    std::vector<int16_t> res;
    int resSize = fastbit_get_result_rows( q->second ) ;

    if ( resSize == 0 )
      return;

    for ( int i = 0; i < resSize; i++ ) {
        res.push_back( *qresult );
        qresult++;
    }

    Local<Array> array = Array::New( isolate, resSize );

    uint32_t idx = 0;
    for (std::vector<int16_t>::iterator it = res.begin(); it != res.end(); ++it) {
        array->Set( idx, Integer::New(isolate, *it) );
        idx++;
    }
    args.GetReturnValue().Set( array );
}

void NodeFastbit::get_qualified_ints( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator q;
    q = obj->m_queryHandles.find( queryHandle );
    if (q == obj->m_queryHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );
    const int32_t *qresult =
        fastbit_get_qualified_ints( q->second, cname );

    if ( qresult == 0 )
      return;

    std::vector<int32_t> res;
    int resSize = fastbit_get_result_rows( q->second ) ;

    if ( resSize == 0 )
      return;

    for ( int i = 0; i < resSize; i++ ) {
        res.push_back( *qresult );
        qresult++;
    }

    Local<Array> array = Array::New( isolate, resSize );

    uint32_t idx = 0;
    for (std::vector<int32_t>::iterator it = res.begin(); it != res.end(); ++it) {
        array->Set( idx, Integer::New(isolate, *it) );
        idx++;
    }
    args.GetReturnValue().Set( array );
}

void NodeFastbit::get_qualified_longs( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator q;
    q = obj->m_queryHandles.find( queryHandle );
    if (q == obj->m_queryHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );
    const int64_t *qresult = 
        fastbit_get_qualified_longs( q->second, cname );

    if ( qresult == 0 )
      return;

    std::vector<int64_t> res;
    int resSize = fastbit_get_result_rows( q->second ) ;

    if ( resSize == 0 )
      return;

    for ( int i = 0; i < resSize; i++ ) {
        res.push_back( *qresult );
        qresult++;
    }

    Local<Array> array = Array::New( isolate, resSize );

    uint32_t idx = 0;
    for (std::vector<int64_t>::iterator it = res.begin(); it != res.end(); ++it) {
        array->Set( idx, Integer::New(isolate, *it) );
        idx++;
    }
    args.GetReturnValue().Set( array );
}

void NodeFastbit::get_qualified_ubytes( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator q;
    q = obj->m_queryHandles.find( queryHandle );
    if (q == obj->m_queryHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );
    const unsigned char* qresult =
        fastbit_get_qualified_ubytes( q->second, cname );

    if ( qresult == 0 )
      return;

    std::vector<const unsigned char*> res;
    int resSize = fastbit_get_result_rows( q->second ) ;

    if ( resSize == 0 )
      return;

    for ( int i = 0; i < resSize; i++ ) {
        res.push_back( qresult );
        qresult++;
    }

    Local<Array> array = Array::New( isolate, resSize );

    uint32_t idx = 0;
    for (std::vector<const unsigned char*>::iterator it = res.begin();
         it != res.end(); ++it) {
        std::ostringstream ss;
        ss << *it;
        array->Set( idx, String::NewFromUtf8( isolate, ss.str().c_str() ) );
        idx++;
    }
    args.GetReturnValue().Set( array );
}

void NodeFastbit::get_qualified_ushorts( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator q;
    q = obj->m_queryHandles.find( queryHandle );
    if (q == obj->m_queryHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );
    const uint16_t *qresult =
      fastbit_get_qualified_ushorts( q->second, cname );

    if ( qresult == 0 )
      return;

    std::vector<uint16_t> res;
    int resSize = fastbit_get_result_rows( q->second ) ;

    if ( resSize == 0 )
      return;

    for ( int i = 0; i < resSize; i++ ) {
        res.push_back( *qresult );
        qresult++;
    }

    Local<Array> array = Array::New( isolate, resSize );

    uint32_t idx = 0;
    for (std::vector<uint16_t>::iterator it = res.begin(); it != res.end(); ++it) {
        array->Set( idx, Integer::New(isolate, *it) );
        idx++;
    }
    args.GetReturnValue().Set( array );
}

void NodeFastbit::get_qualified_uints( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator q;
    q = obj->m_queryHandles.find( queryHandle );
    if (q == obj->m_queryHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );
    const uint32_t *qresult =
      fastbit_get_qualified_uints( q->second, cname );

    if ( qresult == 0 )
      return;

    std::vector<uint32_t> res;
    int resSize = fastbit_get_result_rows( q->second ) ;

    if ( resSize == 0 )
      return;

    for ( int i = 0; i < resSize; i++ ) {
        res.push_back( *qresult );
        qresult++;
    }

    Local<Array> array = Array::New( isolate, resSize );

    uint32_t idx = 0;
    for (std::vector<uint32_t>::iterator it = res.begin(); it != res.end(); ++it) {
        array->Set( idx, Integer::New(isolate, *it) );
        idx++;
    }
    args.GetReturnValue().Set( array );
}

void NodeFastbit::get_qualified_ulongs( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator q;
    q = obj->m_queryHandles.find( queryHandle );
    if (q == obj->m_queryHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );
    const uint64_t *qresult =
      fastbit_get_qualified_ulongs( q->second, cname );

    if ( qresult == 0 )
      return;

    std::vector<uint64_t> res;
    int resSize = fastbit_get_result_rows( q->second ) ;

    if ( resSize == 0 )
      return;

    for ( int i = 0; i < resSize; i++ ) {
        res.push_back( *qresult );
        qresult++;
    }

    Local<Array> array = Array::New( isolate, resSize );

    uint32_t idx = 0;
    for (std::vector<uint64_t>::iterator it = res.begin(); it != res.end(); ++it) {
        array->Set( idx, Integer::New(isolate, *it) );
        idx++;
    }
    args.GetReturnValue().Set( array );
}

void NodeFastbit::get_qualified_strings( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator q;
    q = obj->m_queryHandles.find( queryHandle );
    if (q == obj->m_queryHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );
    const char* qresult = 
        *fastbit_get_qualified_strings( q->second, cname ); // ?

    if ( qresult == 0 )
      return;

    std::vector<const char*> res;
    int resSize = fastbit_get_result_rows( q->second ) ;

    if ( resSize == 0 )
      return;

    for ( int i = 0; i < resSize; i++ ) {
        res.push_back( qresult );
        qresult++;
    }

    Local<Array> array = Array::New( isolate, resSize );

    uint32_t idx = 0;
    for (std::vector<const char*>::iterator it = res.begin();
         it != res.end(); ++it) {
        std::ostringstream ss;
        ss << *it;
        array->Set( idx, String::NewFromUtf8( isolate, ss.str().c_str() ) );
        idx++;
    }
    args.GetReturnValue().Set( array );
}

void NodeFastbit::build_result_set( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string queryHandle( *String::Utf8Value( args[0]->ToString() ) );
    std::string resSetHandle( *String::Utf8Value( args[1]->ToString() ) );

    std::map<std::string,FastBitQuery*>::iterator q;
    q = obj->m_queryHandles.find( queryHandle );
    if (q == obj->m_queryHandles.end())
      return;

    if (!q->second)
      return;

    FastBitResultSet* resultSet = fastbit_build_result_set( q->second );
    std::pair< std::string, FastBitResultSet* > builtSet( resSetHandle,
            resultSet );
    obj->m_resultSetHandles.insert( builtSet );
    args.GetReturnValue().Set( args[1]->ToString() );
}

void NodeFastbit::destroy_result_set( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    // custom implementation
    // TODO: Check for memory leaks
    obj->m_resultSetHandles.erase( resSetHandle );
    args.GetReturnValue().Set( Integer::New( isolate, 0 ) );
}

void NodeFastbit::result_set_next( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 1 ) ||
         isWrongType( !args[0]->IsString(), 0 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    int result = fastbit_result_set_next( q->second );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::result_set_next_bundle( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    int result = fastbit_result_set_next_bundle( q->second );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::result_set_get_int( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );
    int result = fastbit_result_set_get_int( q->second, cname );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::result_set_get_unsigned( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );

    unsigned result = fastbit_result_set_get_unsigned( q->second, cname );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::result_set_get_long( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );

    int64_t result = fastbit_result_set_get_long( q->second, cname );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::result_set_get_float( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );

    float result = fastbit_result_set_get_float( q->second, cname );
    args.GetReturnValue().Set( Number::New( isolate, result ) );
}

void NodeFastbit::result_set_get_double( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );

    double result = fastbit_result_set_get_double( q->second, cname );
    args.GetReturnValue().Set( Number::New( isolate, result ) );
}

void NodeFastbit::result_set_get_string( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    const char* cname = getChar( args[1] );

    std::ostringstream ss;
    const char* qresult = fastbit_result_set_get_string( q->second, cname );
    ss << qresult;
    args.GetReturnValue().Set( String::NewFromUtf8( isolate, ss.str().c_str() ) );
}

void NodeFastbit::result_set_getInt( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    unsigned pos = args[1]->Uint32Value(); // BEWARE
    int32_t result = fastbit_result_set_getInt( q->second, pos );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::result_set_getUnsigned( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    unsigned pos = args[1]->Uint32Value(); // BEWARE
    uint32_t result = fastbit_result_set_getUnsigned( q->second, pos );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::result_set_getLong( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    unsigned pos = args[1]->Uint32Value(); // BEWARE
    int64_t result = fastbit_result_set_getLong( q->second, pos );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::result_set_getFloat( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    unsigned pos = args[1]->Uint32Value(); // BEWARE
    float result = fastbit_result_set_getFloat( q->second, pos );
    args.GetReturnValue().Set( Number::New( isolate, result ) );
}

void NodeFastbit::result_set_getDouble( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    unsigned pos = args[1]->Uint32Value(); // BEWARE
    double result = fastbit_result_set_getDouble( q->second, pos );
    args.GetReturnValue().Set( Number::New( isolate, result ) );
}

void NodeFastbit::result_set_getString( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 2 ) ) {
      return;
    }

    std::string resSetHandle( *String::Utf8Value( args[0]->ToString() ) );

    std::map<std::string,FastBitResultSet*>::iterator q;
    q = obj->m_resultSetHandles.find( resSetHandle );
    if (q == obj->m_resultSetHandles.end())
      return;

    if (!q->second)
      return;

    unsigned pos = args[1]->Uint32Value(); // BEWARE
    std::ostringstream ss;

    const char *qresult = fastbit_result_set_getString( q->second, pos );
    ss << qresult;
    args.GetReturnValue().Set( String::NewFromUtf8( isolate, ss.str().c_str() ) );
}

void NodeFastbit::flush_buffer( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    const char *dir = getChar( args[0] );
    int result = fastbit_flush_buffer(dir);
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::add_values( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 5 ) ) {
      return;
    }

    int ierr = -1;
    const char *colname = getChar( args[0] );
    const char *coltype = getChar( args[1] );

    // array of values from JS side
    v8::Handle<v8::Array> array = v8::Handle<v8::Array>::Cast( args[2] );

    // TRICKY: see capi.cpp line 2567
    void *vals;

    std::vector<std::string>   bytesnames; ///!< For ibis::BYTE.
    std::vector<signed char>   bytesvalues;
    std::vector<std::string>   ubytesnames; ///!< For ibis::UBYTE.
    std::vector<unsigned char> ubytesvalues;
    std::vector<std::string>   shortsnames; ///!< For ibis::SHORT.
    std::vector<int16_t>       shortsvalues;
    std::vector<std::string>   ushortsnames; ///!< For ibis::USHORT.
    std::vector<uint16_t>      ushortsvalues;
    std::vector<std::string>   intsnames; ///!< For ibis::INT.
    std::vector<int32_t>       intsvalues;
    std::vector<std::string>   uintsnames; ///!< For ibis::UINT.
    std::vector<uint32_t>      uintsvalues;
    std::vector<std::string>   longsnames; ///!< For ibis::LONG.
    std::vector<int64_t>       longsvalues;
    std::vector<std::string>   ulongsnames; ///!< For ibis::ULONG.
    std::vector<uint64_t>      ulongsvalues;
    std::vector<std::string>   floatsnames; ///!< For ibis::FLOAT.
    std::vector<float>         floatsvalues;
    std::vector<std::string>   doublesnames; ///!< For ibis::DOUBLE.
    std::vector<double>        doublesvalues;
    std::vector<std::string>   catsnames; ///!< For ibis::CATEGORY.
    std::vector<std::string>   catsvalues;
    std::vector<std::string>   textsnames; ///!< For ibis::TEXT.
    std::vector<std::string>   textsvalues;
    std::vector<std::string>   blobsnames; ///!< For ibis::BLOB
    std::vector<ibis::opaque>  blobsvalues;

    int result = 0;
    uint32_t nelem = args[3]->Uint32Value();
    uint32_t start = args[4]->Uint32Value();

    // need some function to shove values in.
    if (colname == 0 || coltype == 0) {
        args.GetReturnValue().Set( Integer::New( isolate, ierr ) );
        return;
    }
    while (*colname != 0 && isspace(*colname)) ++ colname;
    while (*coltype != 0 && isspace(*coltype)) ++ coltype;
    if (*colname == 0 || *coltype == 0) {
        args.GetReturnValue().Set( Integer::New( isolate, ierr ) );
        return;
    }
    if (nelem == 0) {
        args.GetReturnValue().Set( Integer::New( isolate, 0 ) );
        return;
    }

    ierr = 0;
    switch (*coltype) {
    default : result = -2; break;
    case 'D':
    case 'd': {
          for ( unsigned int i = 0; i < nelem; i++ ) {
            doublesvalues.push_back( array->Get(i)->NumberValue() );
          }
          vals = doublesvalues.data();
          }; break;
    case 'F':
    case 'f': {
          for ( unsigned int i = 0; i < nelem; i++ ) {
            floatsvalues.push_back( (float)array->Get(i)->NumberValue() );
          }
          vals = floatsvalues.data();
              }; break;
    case 'L':
    case 'l': {
          for ( unsigned int i = 0; i < nelem; i++ ) {
              longsvalues.push_back( array->Get(i)->Int32Value() );
          }
          vals = longsvalues.data();
              }; break;
    case 'I':
    case 'i': {
          for ( unsigned int i = 0; i < nelem; i++ ) {
              intsvalues.push_back( array->Get(i)->Int32Value() );
          }
          vals = intsvalues.data();
              }; break;
    case 'S':
    case 's': {
          for ( unsigned int i = 0; i < nelem; i++ ) {
              shortsvalues.push_back( (int16_t)array->Get(i)->Int32Value() );
          }
          vals = shortsvalues.data();
              }; break;
    case 'B':
    case 'b': {
          for ( unsigned int i = 0; i < nelem; i++ ) {
              bytesvalues.push_back( (signed char) *getChar(array->Get(i)) );
          }
          vals = bytesvalues.data();
              }; break;
    case 'C':
    case 'c':
    case 'K':
    case 'k': {// category
          for ( unsigned int i = 0; i < nelem; i++ ) {
              std::string kat(  *String::Utf8Value( array->Get(i)->ToString() ) );
              catsvalues.push_back( kat );
          }
          vals = catsvalues.data();
              }; break;
    case 'T':
    case 't': {
          for ( unsigned int i = 0; i < nelem; i++ ) {
              std::string text(  *String::Utf8Value( array->Get(i)->ToString() ) );
              textsvalues.push_back( text );
          }
          vals = textsvalues.data();
              }; break;
    case 'U':
    case 'u': {
        switch (coltype[1]) {
        default : ierr = -2; break;
            case 'L':
            case 'l': {
                  for ( unsigned int i = 0; i < nelem; i++ ) {
                      // BEWARE: v8 only goes up to int64_t
                      ulongsvalues.push_back( (uint64_t)array->Get(i)->IntegerValue() );
                  }
                  vals = ulongsvalues.data();
                      }; break;
            case 'I':
            case 'i': {
                  for ( unsigned int i = 0; i < nelem; i++ ) {
                      uintsvalues.push_back( array->Get(i)->Uint32Value() );
                  }
                  vals = uintsvalues.data();
                      }; break;
            case 'S':
            case 's': {
                  for ( unsigned int i = 0; i < nelem; i++ ) {
                      ushortsvalues.push_back( (uint16_t)array->Get(i)->Uint32Value() );
                  }
                  vals = ushortsvalues.data();
                      }; break;
            case 'B':
            case 'b': {
                  for ( unsigned int i = 0; i < nelem; i++ ) {
                      ubytesvalues.push_back( (unsigned char) *getChar(array->Get(i)) );
                  }
                  vals = ubytesvalues.data();
                      }; break;
        }
        break;}
    }

    result = fastbit_add_values( colname, coltype, vals, nelem, start);
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::rows_in_partition( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    const char *dir = getChar( args[0] );
    int result = fastbit_rows_in_partition(dir);
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::columns_in_partition( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    const char *dir = getChar( args[0] );
    int result = fastbit_columns_in_partition(dir);
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::init( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    const char *rcfile = getChar( args[0] );
    fastbit_init( rcfile );
}

void NodeFastbit::cleanup( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 0 ) ) {
      return;
    }
    fastbit_cleanup(); // TODO: Does this need to have the queries defined in
                       //       same scope?
}

void NodeFastbit::set_verbose_level( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    int verboseLevel = (int)args[0]->IntegerValue(); // WARN: cast
    int result = fastbit_set_verbose_level( verboseLevel );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::get_verbose_level( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 0 ) ) {
      return;
    }

    int result = fastbit_get_verbose_level();
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::set_logfile( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 1 ) ) {
      return;
    }

    const char *filename = getChar( args[0] );
    int result = fastbit_set_logfile( filename );
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

void NodeFastbit::get_logfile( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 0 ) ) {
      return;
    }

    std::ostringstream ss;
    const char* qresult = fastbit_get_logfile();
    ss << qresult;
    args.GetReturnValue().Set( String::NewFromUtf8( isolate, ss.str().c_str() ) );
}

void NodeFastbit::get_logfilepointer( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );
    NodeFastbit* obj = ObjectWrap::Unwrap<NodeFastbit>( args.Holder() );

    if ( isWrongArgNumber( args, 0 ) ) {
      return;
    }

    obj->m_logFileHandle = fastbit_get_logfilepointer();
}

void NodeFastbit::read_clock( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 0 ) ) {
      return;
    }

    double clock = fastbit_read_clock();
    args.GetReturnValue().Set( Number::New( isolate, clock ) );
}

void NodeFastbit::get_version_string( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 0 ) ) {
      return;
    }

    std::ostringstream ss;
    const char* qresult = fastbit_get_version_string();
    ss << qresult;
    args.GetReturnValue().Set( String::NewFromUtf8( isolate, ss.str().c_str() ) );
}

void NodeFastbit::get_version_number( const FunctionCallbackInfo<Value>& args ) {
    Isolate* isolate = Isolate::GetCurrent();
    HandleScope scope( isolate );

    if ( isWrongArgNumber( args, 0 ) ) {
      return;
    }

    int result = fastbit_get_version_number();
    args.GetReturnValue().Set( Integer::New( isolate, result ) );
}

//---- helper functions 
char *NodeFastbit::getChar( v8::Local<v8::Value> value, const char *fallback ) {
    if ( value->IsString() ) {
        v8::String::Utf8Value string( value );
        char *str  = ( char * ) malloc( string.length() + 1 );
        std::strcpy( str, *string );
        return str;
    }
    char *str = ( char * ) malloc( std::strlen( fallback ) + 1 );
    std::strcpy( str, fallback );
    return str;
}

bool NodeFastbit::isWrongArgNumber( const FunctionCallbackInfo<Value> &args,
                                  int argNumExpected ) {
    Isolate* isolate = Isolate::GetCurrent();
    if ( args.Length() != argNumExpected ) {
        isolate->ThrowException(
            Exception::TypeError(
                String::NewFromUtf8(isolate, "Wrong number of arguments" ) ) );
        return true;
    }

    for (int ii = 0; ii < args.Length(); ii++) {
        if ( args[ii]->IsUndefined() ) {
            std::ostringstream ss;
            if (ii == 0) {
                ss << "1st";
            }
            else if (ii == 1) {
                ss << "2nd";
            }
            else if (ii == 2) {
                ss << "3rd";
            }
            else if (ii > 2) {
                ss << ii+1 << "th";
            }
            ss << " argument is undefined.";

            isolate->ThrowException( Exception::TypeError(
                String::NewFromUtf8(isolate, ss.str().c_str() ) ) );
            return true;
        }
    }
    return false;
}

bool NodeFastbit::isWrongType( bool predicateRes, int argId ) {
    Isolate* isolate = Isolate::GetCurrent();
    if ( predicateRes ) {
        std::ostringstream ss;
        ss << "Argument " << argId << " is of wrong type.";
        isolate->ThrowException(
            Exception::TypeError(
                String::NewFromUtf8(isolate, ss.str().c_str() ) ) );
        return true;
    }
    return false;
}
