'use strict';
const fs = require('fs-extra');
const fastbit = require('../fastbit.js');
const expect = require('expect');
const uuid = require('uuid');
const csv = require('csv');

fastbit.set_verbose_level(0);

const transposeData = (columnHeaders, data) => {
  let transposedData = new Array(columnHeaders.length)
  columnHeaders.forEach((header, col) => {
    transposedData[col] = [];
  })
  data.forEach((item) => {
    columnHeaders.forEach((header, col) => {
      transposedData[col].push(item[col]);
    })
  })
  return transposedData;
}

let testSetData;

// Alternatively, data can be prepped with
// ardea -d testdata -m "a:int, b:float, c:short" -t tests/test0.csv
beforeEach( function (done) {
  fs.readFile('./test/test0.csv', function(err, data){
    if (err) return done(err);
    csv.parse(data, function(err, data){
      if (err) return done(err);
      csv.transform(data, function(data){
        return data.map(function(value){return parseFloat(value.trim())});
      }, function(err, data){
        if (err) return done(err);
        let columnHeaders = ['a', 'b', 'c'];
        testSetData = transposeData(columnHeaders, data);
        return done()
      });
    });
  });
});

afterEach( function (done) {
  fs.remove('./test/testdata', (err, data) => {
    if (err) return done(err);
    return done();
  })
});

describe('index operations', function () {
  beforeEach( function () {
    let intArray = new Int32Array(testSetData[0])
    let shortArray = new Int16Array(testSetData[2])
    fastbit.add_values('a', 'int',  intArray, testSetData[0].length, 0);
    fastbit.add_values('b', 'float', testSetData[1], testSetData[1].length, 0);
    fastbit.add_values('c', 'short',  shortArray, testSetData[2].length, 0);
    fastbit.flush_buffer("./test/testdata");
    expect( fastbit.build_index( './test/testdata', 'a', 'none') ).toEqual(0);
    expect( fastbit.build_index( './test/testdata', 'b', 'none') ).toEqual(0);
    expect( fastbit.build_index( './test/testdata', 'c', 'none') ).toEqual(0);
  });

  it( 'should build indexes correctly', function () {
    // TODO figure out index options in build_indexes()
    expect( fastbit.build_indexes( './test/testdata', 'none') ).toEqual(0);
  });

  it( 'should build individual indices correctly', function () {
    expect( fastbit.build_index( './test/testdata', 'a', 'none') ).toEqual(0);
    expect( fastbit.build_index( './test/testdata', 'b', 'none') ).toEqual(0);
    expect( fastbit.build_index( './test/testdata', 'c', 'none') ).toEqual(0);
  });

  it( 'should purge indexes correctly', function () {
    expect( fastbit.build_indexes( './test/testdata', 'none') ).toEqual(0);
    expect( fastbit.purge_indexes( './test/testdata' ) ).toEqual(0);
  });

  it( 'should return error code of 1 when data directory does not exist', function () {
    // expect( fastbit.purge_indexes( './test/foobar' ) ).toEqual(1);
  });

  it( 'should purge individual indices correctly', function () {
    expect( fastbit.build_indexes( './test/testdata', 'none') ).toEqual(0);
    expect( fastbit.purge_index( './test/testdata', 'a' ) ).toEqual(0);
    expect( fastbit.purge_index( './test/testdata', 'b' ) ).toEqual(0);
    expect( fastbit.purge_index( './test/testdata', 'c' ) ).toEqual(0);
    expect( fastbit.purge_index( './test/testdata', 'foobar' ) ).toEqual(-2);
    // expect( fastbit.purge_index( './test/foobar', 'foobar' ) ).toEqual(1);
  });

  it( 'should reorder partition', function () {
    // this gets stuck because trylock fails
    // expect( fastbit.reorder_partition( './test/testdata' ).toEqual(0) );
  });
});

describe('loading data', function () {
  beforeEach( function () {
    fastbit.add_values('a', 'int',  testSetData[0], testSetData[0].length, 0);
    fastbit.add_values('b', 'float', testSetData[1], testSetData[1].length, 0);
    fastbit.add_values('c', 'short',  testSetData[2], testSetData[2].length, 0);
    fastbit.flush_buffer("./test/testdata");
  });

  // check to see if loaded data reflects csv data entry
  it( 'should have 100 rows', function () {
    let numRowTotal = fastbit.rows_in_partition('./test/testdata');
    expect( numRowTotal ).toEqual( 100 );
  });

  it( 'should have 3 columns', function () {
    let numColTotal = fastbit.columns_in_partition('./test/testdata');
    expect( numColTotal ).toEqual( 3 );
  });
  it( 'should have [99,0] to have value of 99', function () {
    let queryHandle = uuid.v4();
    fastbit.build_query( queryHandle, 'a', "test/testdata", "a = 99" );
    let rowsInA = fastbit.get_result_rows( queryHandle );
    let rowRes = fastbit.get_result_row_ids( queryHandle );

    expect( rowsInA ).toEqual( 1 );
    expect( rowRes.ret ).toEqual( 1 );
    expect( rowRes.ids ).toEqual( [99] );

    fastbit.destroy_query( queryHandle );
  });
});

describe('query functionalities', function () {
  let queryHandle = uuid.v4();

  beforeEach( function () {
    let allAChars = []
    for (let i = 0; i < testSetData[0].length; i++) {
      allAChars.push('A')
    }
    allAChars[90] = 'B'
    allAChars[91] = 'C'
    allAChars[92] = 'D'
    allAChars[93] = 'E'
    let stringArr = []
    for (let i = 0; i < testSetData[0].length; i++) {
      stringArr.push('foo')
    }
    stringArr[91] = 'bif'
    stringArr[92] = 'zip'
    stringArr[93] = 'moo'
    stringArr[94] = 'bar'
    fastbit.add_values('int', 'int',  testSetData[0], testSetData[0].length, 0);
    fastbit.add_values('float', 'float', testSetData[1], testSetData[1].length, 0);
    fastbit.add_values('short', 'short',  testSetData[2], testSetData[2].length, 0);
    fastbit.add_values('byte', 'byte', allAChars, allAChars.length, 0)
    fastbit.add_values('double', 'double', testSetData[1], testSetData[1].length, 0);
    fastbit.add_values('long', 'long',  testSetData[0], testSetData[0].length, 0);
    fastbit.add_values('ubyte', 'ubyte', allAChars, allAChars.length, 0)
    fastbit.add_values('ushort', 'ushort',  testSetData[2], testSetData[2].length, 0);
    fastbit.add_values('uint', 'uint',  testSetData[0], testSetData[0].length, 0);
    fastbit.add_values('ulong', 'ulong',  testSetData[0], testSetData[0].length, 0);
    fastbit.add_values('string', 'string', stringArr, stringArr.length, 0)

    fastbit.flush_buffer("./test/testdata");
    expect( fastbit.build_query( queryHandle,
      'int, float, short, byte, double, long, ubyte, ushort, uint, ulong, string',
      'test/testdata', 'int > 90 AND int < 98 AND float < 95' ) ).toEqual(queryHandle)
  });

  afterEach( function () {
    // TODO add tests for other error codes
    expect( fastbit.destroy_query( queryHandle ) ).toEqual(0);
  });

  it( 'should get correct numbers of result columns', function () {
    expect( fastbit.get_result_columns( queryHandle ) ).toEqual(11);
  });

  it( 'should get correct numbers of result rows', function () {
    expect( fastbit.get_result_rows( queryHandle ) ).toEqual(4);
  });

  it( 'should get correct row ids', function () {
    expect( fastbit.get_result_row_ids( queryHandle ) )
      .toEqual({ids:[91, 92, 93, 94], ret:4});
  });

  it( 'should return select clause', function () {
    expect( fastbit.get_select_clause( queryHandle ) )
      .toEqual('int, float, short, byte, double, long, ubyte, ushort, uint, ulong, string');
  });

  it( 'should return from clause', function () {
    expect( fastbit.get_from_clause( queryHandle ) ).toEqual('testdata');
  });

  it( 'should return where clause', function () {
    expect( fastbit.get_where_clause( queryHandle ) ).toEqual('int > 90 AND int < 98 AND float < 95');
  });

  it( 'should return qualified floats from query', function () {
    expect( fastbit.get_qualified_floats( queryHandle, 'int' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_floats( queryHandle, 'float' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_floats( queryHandle, 'short' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_floats( queryHandle, 'byte' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_floats( queryHandle, 'double' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_floats( queryHandle, 'long' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_floats( queryHandle, 'ubyte' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_floats( queryHandle, 'ushort' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_floats( queryHandle, 'uint' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_floats( queryHandle, 'ulong' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_floats( queryHandle, 'string' ) ).toEqual(undefined)
  });

  it( 'should return qualified doubles from query', function () {
    expect( fastbit.get_qualified_doubles( queryHandle, 'int' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_doubles( queryHandle, 'float' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_doubles( queryHandle, 'short' ) ).toEqual([ 8, 7, 6, 5 ])
    expect( fastbit.get_qualified_doubles( queryHandle, 'byte' ) ).toEqual([ 67, 68, 69, 65 ])
    expect( fastbit.get_qualified_doubles( queryHandle, 'double' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_doubles( queryHandle, 'long' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_doubles( queryHandle, 'ubyte' ) ).toEqual([ 67, 68, 69, 65 ])
    expect( fastbit.get_qualified_doubles( queryHandle, 'ushort' ) ).toEqual([ 8, 7, 6, 5 ])
    expect( fastbit.get_qualified_doubles( queryHandle, 'uint' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_doubles( queryHandle, 'ulong' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_doubles( queryHandle, 'string' ) ).toEqual([ 0, 0, 0, 0 ]) // ?
  });

  it( 'should return qualified bytes from query', function () {
    expect( fastbit.get_qualified_bytes( queryHandle, 'int' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'float' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'short' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'byte' ) ).toEqual(['CDEA', 'DEA', 'EA', 'A' ])
    expect( fastbit.get_qualified_bytes( queryHandle, 'double' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'long' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'ubyte' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'ushort' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'uint' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'ulong' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'string' ) ).toEqual(undefined)
  });

  it( 'should return qualified shorts from query', function () {
    expect( fastbit.get_qualified_shorts( queryHandle, 'int' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_shorts( queryHandle, 'float' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_shorts( queryHandle, 'short' ) ).toEqual([ 8, 7, 6, 5 ])
    expect( fastbit.get_qualified_shorts( queryHandle, 'byte' ) ).toEqual([ 67, 68, 69, 65 ])
    expect( fastbit.get_qualified_shorts( queryHandle, 'double' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_shorts( queryHandle, 'long' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_shorts( queryHandle, 'ubyte' ) ).toEqual([ 67, 68, 69, 65 ])
    expect( fastbit.get_qualified_shorts( queryHandle, 'ushort' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_shorts( queryHandle, 'uint' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_shorts( queryHandle, 'ulong' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_shorts( queryHandle, 'string' ) ).toEqual([ 0, 0, 0, 0 ])
  });

  it( 'should return qualified ints from query', function () {
    expect( fastbit.get_qualified_ints( queryHandle, 'int' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_ints( queryHandle, 'float' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ints( queryHandle, 'short' ) ).toEqual([ 8, 7, 6, 5 ])
    expect( fastbit.get_qualified_ints( queryHandle, 'byte' ) ).toEqual([ 67, 68, 69, 65 ])
    expect( fastbit.get_qualified_ints( queryHandle, 'double' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ints( queryHandle, 'long' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ints( queryHandle, 'ubyte' ) ).toEqual([ 67, 68, 69, 65 ])
    expect( fastbit.get_qualified_ints( queryHandle, 'ushort' ) ).toEqual([ 8, 7, 6, 5 ])
    expect( fastbit.get_qualified_ints( queryHandle, 'uint' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ints( queryHandle, 'ulong' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ints( queryHandle, 'string' ) ).toEqual([ 0, 0, 0, 0 ])
  });

  it( 'should return qualified longs from query', function () {
    expect( fastbit.get_qualified_longs( queryHandle, 'int' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_longs( queryHandle, 'float' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_longs( queryHandle, 'short' ) ).toEqual([ 8, 7, 6, 5 ])
    expect( fastbit.get_qualified_longs( queryHandle, 'byte' ) ).toEqual([ 67, 68, 69, 65 ])
    expect( fastbit.get_qualified_longs( queryHandle, 'double' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_longs( queryHandle, 'long' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_longs( queryHandle, 'ubyte' ) ).toEqual([ 67, 68, 69, 65 ])
    expect( fastbit.get_qualified_longs( queryHandle, 'ushort' ) ).toEqual([ 8, 7, 6, 5 ])
    expect( fastbit.get_qualified_longs( queryHandle, 'uint' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_longs( queryHandle, 'ulong' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_longs( queryHandle, 'string' ) ).toEqual([ 0, 0, 0, 0 ])
  });

  // FIXME
  it( 'should return qualified ubytes from query', function () {
    expect( fastbit.get_qualified_bytes( queryHandle, 'int' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'float' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'short' ) ).toEqual(undefined)
    //expect( fastbit.get_qualified_bytes( queryHandle, 'byte' ) ).toEqual(['CDEA', 'DEA', 'EA', 'A' ]) //
    expect( fastbit.get_qualified_bytes( queryHandle, 'double' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'long' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'ubyte' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'ushort' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'uint' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'ulong' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_bytes( queryHandle, 'string' ) ).toEqual(undefined)
  });

  it( 'should return qualified ushorts from query', function () {
    // try columns with negative numbers
    expect( fastbit.get_qualified_ushorts( queryHandle, 'int' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ushorts( queryHandle, 'float' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ushorts( queryHandle, 'short' ) ).toEqual(undefined) // ?
    expect( fastbit.get_qualified_ushorts( queryHandle, 'byte' ) ).toEqual([ 67, 68, 69, 65 ])
    expect( fastbit.get_qualified_ushorts( queryHandle, 'double' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ushorts( queryHandle, 'long' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ushorts( queryHandle, 'ubyte' ) ).toEqual([ 67, 68, 69, 65 ])
    expect( fastbit.get_qualified_ushorts( queryHandle, 'ushort' ) ).toEqual([ 8, 7, 6, 5 ])
    expect( fastbit.get_qualified_ushorts( queryHandle, 'uint' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ushorts( queryHandle, 'ulong' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ushorts( queryHandle, 'string' ) ).toEqual(undefined)
  });

  it( 'should return qualified uints from query', function () {
    // try columns with negative numbers
    expect( fastbit.get_qualified_uints( queryHandle, 'int' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_uints( queryHandle, 'float' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_uints( queryHandle, 'short' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_uints( queryHandle, 'byte' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_uints( queryHandle, 'double' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_uints( queryHandle, 'long' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_uints( queryHandle, 'ubyte' ) ).toEqual([ 67, 68, 69, 65 ])
    expect( fastbit.get_qualified_uints( queryHandle, 'ushort' ) ).toEqual([ 8, 7, 6, 5 ])
    expect( fastbit.get_qualified_uints( queryHandle, 'uint' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_uints( queryHandle, 'ulong' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_uints( queryHandle, 'string' ) ).toEqual(undefined)
  });

  it( 'should return qualified ulongs from query', function () {
    // try columns with negative numbers
    expect( fastbit.get_qualified_ulongs( queryHandle, 'int' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ulongs( queryHandle, 'float' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ulongs( queryHandle, 'short' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ulongs( queryHandle, 'byte' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ulongs( queryHandle, 'double' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ulongs( queryHandle, 'long' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_ulongs( queryHandle, 'ubyte' ) ).toEqual([ 67, 68, 69, 65 ])
    expect( fastbit.get_qualified_ulongs( queryHandle, 'ushort' ) ).toEqual([ 8, 7, 6, 5 ])
    expect( fastbit.get_qualified_ulongs( queryHandle, 'uint' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_ulongs( queryHandle, 'ulong' ) ).toEqual([ 91, 92, 93, 94 ])
    expect( fastbit.get_qualified_ulongs( queryHandle, 'string' ) ).toEqual(undefined)
  });

  // FIXME string should return correct results
  // results show item being overwritten - probably due to incorrect handling of
  // char array
  it.skip( 'should return qualified strings from query', function () {
    expect( fastbit.get_qualified_strings( queryHandle, 'int' ) ).toEqual([ '91', '1', '', '' ])
    expect( fastbit.get_qualified_strings( queryHandle, 'float' ) ).toEqual([ '91', '1', '', '' ])
    expect( fastbit.get_qualified_strings( queryHandle, 'short' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_strings( queryHandle, 'byte' ) ).toEqual(['C', 'D', 'E', 'A' ])
    expect( fastbit.get_qualified_strings( queryHandle, 'double' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_strings( queryHandle, 'long' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_strings( queryHandle, 'ubyte' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_strings( queryHandle, 'ushort' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_strings( queryHandle, 'uint' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_strings( queryHandle, 'ulong' ) ).toEqual(undefined)
    expect( fastbit.get_qualified_strings( queryHandle, 'string' ) ).toEqual([ '91', '1', '', '' ])
  });
});

describe('result set control functionalities', function () {
  beforeEach( function () {
    fastbit.add_values('a', 'int',  testSetData[0], testSetData[0].length, 0)
    fastbit.add_values('b', 'float', testSetData[1], testSetData[1].length, 0)
    fastbit.add_values('c', 'short',  testSetData[2], testSetData[2].length, 0)
    fastbit.flush_buffer("./test/testdata")
  })

  // TODO current function does not throw an error on JS side
  it.skip( 'should fail to build a result set if there is no query', function () {
    let resultHandle = uuid.v4();
    expect(fastbit.build_result_set(null, resultHandle)).toEqual(undefined)
  })

  // TODO current function does not throw an error on JS side
  it.skip( 'should fail to build a result set if there is no result handle', function () {
    let queryHandle = uuid.v4()
    fastbit.build_query( queryHandle, 'b', "test/testdata", "b = 99" )
    expect(fastbit.build_result_set(queryHandle, null)).toEqual(undefined)
    fastbit.destroy_query( queryHandle )
  })

  // TODO current function does not throw an error on JS side
  it.skip( 'should fail to build a result set if there is no valid query given query id', function () {
    let queryHandle = uuid.v4()
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(undefined)
  })

  it( 'should successfully build a result set if query exists', function () {
    let queryHandle = uuid.v4()
    let resultHandle = uuid.v4()
    fastbit.build_query( queryHandle, 'b', "test/testdata", "b = 99" )
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    fastbit.destroy_query( queryHandle )
  })

  it( 'should fail to destroy result set if handle is invalid', function () {
    let queryHandle = uuid.v4()
    fastbit.build_query( queryHandle, 'b', "test/testdata", "b = 99" )
    expect(fastbit.destroy_result_set(null)).toEqual(undefined)
    fastbit.destroy_query( queryHandle )
  })

  it( 'should successfully destroy result set if handle is valid', function () {
    let queryHandle = uuid.v4()
    let resultHandle = uuid.v4()
    fastbit.build_query( queryHandle, 'b', "test/testdata", "b = 99" )
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
    fastbit.destroy_query( queryHandle )
  })

  it( 'should return an error if result handle is string', function () {
    let queryHandle = uuid.v4()
    fastbit.build_query( queryHandle, 'b', "test/testdata", "b > 92" )
    try {
      fastbit.result_set_next(null)
    } catch (e) {
      expect(e instanceof TypeError).toEqual(true)
    }
    try {
      fastbit.result_set_next(1)
    } catch (e) {
      expect(e instanceof TypeError).toEqual(true)
    }
    fastbit.destroy_query( queryHandle )
  })

  it( 'should fail to return next set of result if result set is non-existent', function () {
    let queryHandle = uuid.v4()
    let resultHandle = uuid.v4()
    fastbit.build_query( queryHandle, 'b', "test/testdata", "b > 92" )
    expect(fastbit.result_set_next(resultHandle)).toEqual(undefined)
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(undefined)
    fastbit.destroy_query( queryHandle )
  })

  it( 'should fail to return next set of result if result set is empty', function () {
    let queryHandle = uuid.v4()
    let resultHandle = uuid.v4()
    fastbit.build_query( queryHandle, 'b', "test/testdata", "b > 9000" )
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(-1)
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
    fastbit.destroy_query( queryHandle )
  })

  it( 'should return next set of result correctly if exists', function () {
    let queryHandle = uuid.v4()
    let resultHandle = uuid.v4()
    fastbit.build_query( queryHandle, 'a, b', "test/testdata", "b > 96" )
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_get_int( resultHandle, "a" ) ).toEqual(97);
    expect( fastbit.result_set_get_float( resultHandle, "b" ) ).toEqual(97);
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_get_int( resultHandle, "a" ) ).toEqual(98);
    expect( fastbit.result_set_get_float( resultHandle, "b" ) ).toEqual(98);
    expect(fastbit.result_set_next(resultHandle)).toEqual(-1)
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
    fastbit.destroy_query( queryHandle )
  })

  // TODO how is bundle different from result set iterator?
  it( 'should return next bundle of result sets if exists', function () {
    let queryHandle = uuid.v4()
    let resultHandle = uuid.v4()
    fastbit.build_query( queryHandle, 'b', "test/testdata", "b > 96" )
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next_bundle(resultHandle)).toEqual(0)
    expect( fastbit.result_set_get_float( resultHandle, "b" ) ).toEqual(97);
    expect(fastbit.result_set_next_bundle(resultHandle)).toEqual(0)
    expect( fastbit.result_set_get_float( resultHandle, "b" ) ).toEqual(98);
    expect(fastbit.result_set_next_bundle(resultHandle)).toEqual(-1)
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
    fastbit.destroy_query( queryHandle )
  })

  // camelCase functions return by index, while functions with _ return by cname
  it( 'should return int results from result set using column name', function () {
    let queryHandle = uuid.v4()
    let resultHandle = uuid.v4()
    fastbit.build_query( queryHandle, 'b', "test/testdata", "b > 96" )
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next_bundle(resultHandle)).toEqual(0)
    expect( fastbit.result_set_get_float( resultHandle, "b" ) ).toEqual(97);
    expect(fastbit.result_set_next_bundle(resultHandle)).toEqual(0)
    expect( fastbit.result_set_get_float( resultHandle, "b" ) ).toEqual(98);
    expect(fastbit.result_set_next_bundle(resultHandle)).toEqual(-1)
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
    fastbit.destroy_query( queryHandle )
  })
})
describe('result set get data functionalities', function () {
  let queryHandle;

  beforeEach( function () {
    queryHandle = uuid.v4()
    let allAChars = []
    for (let i = 0; i < testSetData[0].length; i++) {
      allAChars.push('A')
    }
    allAChars[90] = 'B'
    allAChars[91] = 'C'
    allAChars[92] = 'D'
    allAChars[93] = 'E'
    let stringArr = []
    for (let i = 0; i < testSetData[0].length; i++) {
      stringArr.push('foo')
    }
    stringArr[91] = 'bif'
    stringArr[92] = 'zip'
    stringArr[93] = 'moo'
    stringArr[94] = 'bar'
    fastbit.add_values('int', 'int',  testSetData[0], testSetData[0].length, 0);
    fastbit.add_values('float', 'float', testSetData[1], testSetData[1].length, 0);
    fastbit.add_values('short', 'short',  testSetData[2], testSetData[2].length, 0);
    fastbit.add_values('byte', 'byte', allAChars, allAChars.length, 0)
    fastbit.add_values('double', 'double', testSetData[1], testSetData[1].length, 0);
    fastbit.add_values('long', 'long',  testSetData[0], testSetData[0].length, 0);
    fastbit.add_values('ubyte', 'ubyte', allAChars, allAChars.length, 0)
    fastbit.add_values('ushort', 'ushort',  testSetData[2], testSetData[2].length, 0);
    fastbit.add_values('uint', 'uint',  testSetData[0], testSetData[0].length, 0);
    fastbit.add_values('ulong', 'ulong',  testSetData[0], testSetData[0].length, 0);
    fastbit.add_values('string', 'string', stringArr, stringArr.length, 0)

    fastbit.flush_buffer("./test/testdata");
    expect( fastbit.build_query( queryHandle,
      'int, float, short, byte, double, long, ubyte, ushort, uint, ulong, string',
      'test/testdata', 'int > 96' ) ).toEqual(queryHandle)
  });
  afterEach( function () {
    expect( fastbit.destroy_query( queryHandle ) ).toEqual(0);
  });
  it( 'should return int results from result set using column name', function () {
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_get_int( resultHandle, 'int' ) ).toEqual(97);
    expect( fastbit.result_set_get_int( resultHandle, 'float' ) ).toEqual(97);
    expect( fastbit.result_set_get_int( resultHandle, 'short' ) ).toEqual(2);
    expect( fastbit.result_set_get_int( resultHandle, 'byte' ) ).toEqual(65);
    expect( fastbit.result_set_get_int( resultHandle, 'double' ) ).toEqual(97);
    expect( fastbit.result_set_get_int( resultHandle, 'long' ) ).toEqual(97);
    expect( fastbit.result_set_get_int( resultHandle, 'ubyte' ) ).toEqual(65);
    expect( fastbit.result_set_get_int( resultHandle, 'ushort' ) ).toEqual(2);
    expect( fastbit.result_set_get_int( resultHandle, 'uint' ) ).toEqual(97);
    expect( fastbit.result_set_get_int( resultHandle, 'ulong' ) ).toEqual(97);
    expect( fastbit.result_set_get_int( resultHandle, 'string' ) ).toEqual(0);
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
  })
  it( 'should return unsigned int results from result set using column name', function () {
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_get_unsigned( resultHandle, 'int' ) ).toEqual(97);
    expect( fastbit.result_set_get_unsigned( resultHandle, 'float' ) ).toEqual(97);
    expect( fastbit.result_set_get_unsigned( resultHandle, 'short' ) ).toEqual(2);
    expect( fastbit.result_set_get_unsigned( resultHandle, 'byte' ) ).toEqual(65);
    expect( fastbit.result_set_get_unsigned( resultHandle, 'double' ) ).toEqual(97);
    expect( fastbit.result_set_get_unsigned( resultHandle, 'long' ) ).toEqual(97);
    expect( fastbit.result_set_get_unsigned( resultHandle, 'ubyte' ) ).toEqual(65);
    expect( fastbit.result_set_get_unsigned( resultHandle, 'ushort' ) ).toEqual(2);
    expect( fastbit.result_set_get_unsigned( resultHandle, 'uint' ) ).toEqual(97);
    expect( fastbit.result_set_get_unsigned( resultHandle, 'ulong' ) ).toEqual(97);
    expect( fastbit.result_set_get_unsigned( resultHandle, 'string' ) ).toEqual(0);
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
  })
  it( 'should return long results from result set using column name', function () {
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_get_long( resultHandle, 'int' ) ).toEqual(97);
    expect( fastbit.result_set_get_long( resultHandle, 'float' ) ).toEqual(97);
    expect( fastbit.result_set_get_long( resultHandle, 'short' ) ).toEqual(2);
    expect( fastbit.result_set_get_long( resultHandle, 'byte' ) ).toEqual(65);
    expect( fastbit.result_set_get_long( resultHandle, 'double' ) ).toEqual(97);
    expect( fastbit.result_set_get_long( resultHandle, 'long' ) ).toEqual(97);
    expect( fastbit.result_set_get_long( resultHandle, 'ubyte' ) ).toEqual(65);
    expect( fastbit.result_set_get_long( resultHandle, 'ushort' ) ).toEqual(2);
    expect( fastbit.result_set_get_long( resultHandle, 'uint' ) ).toEqual(97);
    expect( fastbit.result_set_get_long( resultHandle, 'ulong' ) ).toEqual(97);
    expect( fastbit.result_set_get_long( resultHandle, 'string' ) ).toEqual(0);
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
  })
  it( 'should return float results from result set using column name', function () {
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_get_float( resultHandle, 'int' ) ).toEqual(97);
    expect( fastbit.result_set_get_float( resultHandle, 'float' ) ).toEqual(97);
    expect( fastbit.result_set_get_float( resultHandle, 'short' ) ).toEqual(2);
    expect( fastbit.result_set_get_float( resultHandle, 'byte' ) ).toEqual(65);
    expect( fastbit.result_set_get_float( resultHandle, 'double' ) ).toEqual(97);
    expect( fastbit.result_set_get_float( resultHandle, 'long' ) ).toEqual(97);
    expect( fastbit.result_set_get_float( resultHandle, 'ubyte' ) ).toEqual(65);
    expect( fastbit.result_set_get_float( resultHandle, 'ushort' ) ).toEqual(2);
    expect( fastbit.result_set_get_float( resultHandle, 'uint' ) ).toEqual(97);
    expect( fastbit.result_set_get_float( resultHandle, 'ulong' ) ).toEqual(97);
    expect( fastbit.result_set_get_float( resultHandle, 'string' ) ).toEqual(0);
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
  })
  it( 'should return double results from result set using column name', function () {
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_get_double( resultHandle, 'int' ) ).toEqual(97);
    expect( fastbit.result_set_get_double( resultHandle, 'float' ) ).toEqual(97);
    expect( fastbit.result_set_get_double( resultHandle, 'short' ) ).toEqual(2);
    expect( fastbit.result_set_get_double( resultHandle, 'byte' ) ).toEqual(65);
    expect( fastbit.result_set_get_double( resultHandle, 'double' ) ).toEqual(97);
    expect( fastbit.result_set_get_double( resultHandle, 'long' ) ).toEqual(97);
    expect( fastbit.result_set_get_double( resultHandle, 'ubyte' ) ).toEqual(65);
    expect( fastbit.result_set_get_double( resultHandle, 'ushort' ) ).toEqual(2);
    expect( fastbit.result_set_get_double( resultHandle, 'uint' ) ).toEqual(97);
    expect( fastbit.result_set_get_double( resultHandle, 'ulong' ) ).toEqual(97);
    expect( fastbit.result_set_get_double( resultHandle, 'string' ) ).toEqual(0);
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
  })
  // TODO
  it.skip( 'should return string results from result set using column name', function () {
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_get_string( resultHandle, 'int' ) ).toEqual(97);
    expect( fastbit.result_set_get_string( resultHandle, 'float' ) ).toEqual(97);
    expect( fastbit.result_set_get_string( resultHandle, 'short' ) ).toEqual(2);
    expect( fastbit.result_set_get_string( resultHandle, 'byte' ) ).toEqual(65);
    expect( fastbit.result_set_get_string( resultHandle, 'double' ) ).toEqual(97);
    expect( fastbit.result_set_get_string( resultHandle, 'long' ) ).toEqual(97);
    expect( fastbit.result_set_get_string( resultHandle, 'ubyte' ) ).toEqual(65);
    expect( fastbit.result_set_get_string( resultHandle, 'ushort' ) ).toEqual(2);
    expect( fastbit.result_set_get_string( resultHandle, 'uint' ) ).toEqual(97);
    expect( fastbit.result_set_get_string( resultHandle, 'ulong' ) ).toEqual(97);
    expect( fastbit.result_set_get_string( resultHandle, 'string' ) ).toEqual(0);
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
  })
  it( 'should return int results from result set using column index', function () {
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_getInt( resultHandle, 0 ) ).toEqual(97);
    expect( fastbit.result_set_getInt( resultHandle, 1 ) ).toEqual(97);
    expect( fastbit.result_set_getInt( resultHandle, 2 ) ).toEqual(2);
    expect( fastbit.result_set_getInt( resultHandle, 3 ) ).toEqual(65);
    expect( fastbit.result_set_getInt( resultHandle, 4 ) ).toEqual(97);
    expect( fastbit.result_set_getInt( resultHandle, 5 ) ).toEqual(97);
    expect( fastbit.result_set_getInt( resultHandle, 6 ) ).toEqual(65);
    expect( fastbit.result_set_getInt( resultHandle, 7 ) ).toEqual(2);
    expect( fastbit.result_set_getInt( resultHandle, 8 ) ).toEqual(97);
    expect( fastbit.result_set_getInt( resultHandle, 9 ) ).toEqual(97);
    expect( fastbit.result_set_getInt( resultHandle, 10 ) ).toEqual(0);
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
  })
  it( 'should return unsigned int results from result set using column index', function () {
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_getUnsigned( resultHandle, 0 ) ).toEqual(97);
    expect( fastbit.result_set_getUnsigned( resultHandle, 1 ) ).toEqual(97);
    expect( fastbit.result_set_getUnsigned( resultHandle, 2 ) ).toEqual(2);
    expect( fastbit.result_set_getUnsigned( resultHandle, 3 ) ).toEqual(65);
    expect( fastbit.result_set_getUnsigned( resultHandle, 4 ) ).toEqual(97);
    expect( fastbit.result_set_getUnsigned( resultHandle, 5 ) ).toEqual(97);
    expect( fastbit.result_set_getUnsigned( resultHandle, 6 ) ).toEqual(65);
    expect( fastbit.result_set_getUnsigned( resultHandle, 7 ) ).toEqual(2);
    expect( fastbit.result_set_getUnsigned( resultHandle, 8 ) ).toEqual(97);
    expect( fastbit.result_set_getUnsigned( resultHandle, 9 ) ).toEqual(97);
    expect( fastbit.result_set_getUnsigned( resultHandle, 10 ) ).toEqual(0);
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
  })
  it( 'should return long results from result set using column index', function () {
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_getLong( resultHandle, 0 ) ).toEqual(97);
    expect( fastbit.result_set_getLong( resultHandle, 1 ) ).toEqual(97);
    expect( fastbit.result_set_getLong( resultHandle, 2 ) ).toEqual(2);
    expect( fastbit.result_set_getLong( resultHandle, 3 ) ).toEqual(65);
    expect( fastbit.result_set_getLong( resultHandle, 4 ) ).toEqual(97);
    expect( fastbit.result_set_getLong( resultHandle, 5 ) ).toEqual(97);
    expect( fastbit.result_set_getLong( resultHandle, 6 ) ).toEqual(65);
    expect( fastbit.result_set_getLong( resultHandle, 7 ) ).toEqual(2);
    expect( fastbit.result_set_getLong( resultHandle, 8 ) ).toEqual(97);
    expect( fastbit.result_set_getLong( resultHandle, 9 ) ).toEqual(97);
    expect( fastbit.result_set_getLong( resultHandle, 10 ) ).toEqual(0);
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
  })
  it( 'should return float results from result set using column index', function () {
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_getFloat( resultHandle, 0 ) ).toEqual(97);
    expect( fastbit.result_set_getFloat( resultHandle, 1 ) ).toEqual(97);
    expect( fastbit.result_set_getFloat( resultHandle, 2 ) ).toEqual(2);
    expect( fastbit.result_set_getFloat( resultHandle, 3 ) ).toEqual(65);
    expect( fastbit.result_set_getFloat( resultHandle, 4 ) ).toEqual(97);
    expect( fastbit.result_set_getFloat( resultHandle, 5 ) ).toEqual(97);
    expect( fastbit.result_set_getFloat( resultHandle, 6 ) ).toEqual(65);
    expect( fastbit.result_set_getFloat( resultHandle, 7 ) ).toEqual(2);
    expect( fastbit.result_set_getFloat( resultHandle, 8 ) ).toEqual(97);
    expect( fastbit.result_set_getFloat( resultHandle, 9 ) ).toEqual(97);
    expect( fastbit.result_set_getFloat( resultHandle, 10 ) ).toEqual(0);
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
  })
  it( 'should return double results from result set using column index', function () {
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_getDouble( resultHandle, 0 ) ).toEqual(97);
    expect( fastbit.result_set_getDouble( resultHandle, 1 ) ).toEqual(97);
    expect( fastbit.result_set_getDouble( resultHandle, 2 ) ).toEqual(2);
    expect( fastbit.result_set_getDouble( resultHandle, 3 ) ).toEqual(65);
    expect( fastbit.result_set_getDouble( resultHandle, 4 ) ).toEqual(97);
    expect( fastbit.result_set_getDouble( resultHandle, 5 ) ).toEqual(97);
    expect( fastbit.result_set_getDouble( resultHandle, 6 ) ).toEqual(65);
    expect( fastbit.result_set_getDouble( resultHandle, 7 ) ).toEqual(2);
    expect( fastbit.result_set_getDouble( resultHandle, 8 ) ).toEqual(97);
    expect( fastbit.result_set_getDouble( resultHandle, 9 ) ).toEqual(97);
    expect( fastbit.result_set_getDouble( resultHandle, 10 ) ).toEqual(0);
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
  })
  // TODO
  it.skip( 'should return string results from result set using column index', function () {
    let resultHandle = uuid.v4()
    expect(fastbit.build_result_set(queryHandle, resultHandle)).toEqual(resultHandle)
    expect(fastbit.result_set_next(resultHandle)).toEqual(0)
    expect( fastbit.result_set_getString( resultHandle, 0 ) ).toEqual(97);
    expect( fastbit.result_set_getString( resultHandle, 1 ) ).toEqual(97);
    expect( fastbit.result_set_getString( resultHandle, 2 ) ).toEqual(2);
    expect( fastbit.result_set_getString( resultHandle, 3 ) ).toEqual(65);
    expect( fastbit.result_set_getString( resultHandle, 4 ) ).toEqual(97);
    expect( fastbit.result_set_getString( resultHandle, 5 ) ).toEqual(97);
    expect( fastbit.result_set_getString( resultHandle, 6 ) ).toEqual(65);
    expect( fastbit.result_set_getString( resultHandle, 7 ) ).toEqual(2);
    expect( fastbit.result_set_getString( resultHandle, 8 ) ).toEqual(97);
    expect( fastbit.result_set_getString( resultHandle, 9 ) ).toEqual(97);
    expect( fastbit.result_set_getString( resultHandle, 10 ) ).toEqual(0);
    expect(fastbit.destroy_result_set(resultHandle)).toEqual(0)
  })
})

describe('query functionalities', function () {
  beforeEach( function () {
    fastbit.add_values('a', 'int',  testSetData[0], testSetData[0].length, 0);
    fastbit.add_values('b', 'float', testSetData[1], testSetData[1].length, 0);
    fastbit.add_values('c', 'short',  testSetData[2], testSetData[2].length, 0);
    fastbit.flush_buffer("./test/testdata");
  });

  it( 'should have [99,1] to have no value', function () {
    let queryHandle = uuid.v4();
    fastbit.build_query( queryHandle, 'b', "test/testdata", "b = 99" );
    let rowsInB = fastbit.get_result_rows(queryHandle);
    let rowRes = fastbit.get_result_row_ids( queryHandle );

    expect( rowsInB ).toEqual( 0 );
    expect( !rowRes );

    fastbit.destroy_query( queryHandle );
  });

  it( 'should have [99,2] to have value of 0', function () {
    let queryHandle = uuid.v4();
    fastbit.build_query( queryHandle, 'c', "test/testdata", "c = 0" );
    let rowsInC = fastbit.get_result_rows( queryHandle );
    let rowRes = fastbit.get_result_row_ids( queryHandle );

    expect( rowsInC ).toEqual( 1 );
    expect( rowRes.ret ).toEqual( 1 );
    expect( rowRes.ids ).toEqual( [99] );

    fastbit.destroy_query( queryHandle );
  });

  it( 'should return 9 ids for values in column a greater than 90',
    function () {
    let queryHandle = uuid.v4();
    fastbit.build_query( queryHandle, 'a', "test/testdata", "a > 90" );
    let rowsInA = fastbit.get_result_rows( queryHandle );
    let rowRes = fastbit.get_result_row_ids( queryHandle );

    expect( rowsInA ).toEqual( 9 );
    expect( rowRes.ret ).toEqual( 9 );
    expect( rowRes.ids ).toEqual( [91, 92, 93, 94, 95, 96, 97, 98, 99] );

    fastbit.destroy_query( queryHandle );
  });

  it( 'should return the correct select clause', function () {
    let queryHandle = uuid.v4();
    fastbit.build_query( queryHandle, 'a, b, c', "test/testdata", "a > 90 AND b < 95" );
    expect( fastbit.get_select_clause( queryHandle ) ).toEqual('a, b, c');
    fastbit.destroy_query( queryHandle );
  });

  it( 'should return a value for columns by index', function () {
    let queryHandle = uuid.v4();
    fastbit.build_query( queryHandle, 'a, b, c', "test/testdata", "a > 90 AND b < 95" );

    let resultHandle = uuid.v4();
    fastbit.build_result_set( queryHandle, resultHandle );
    // calling on the getDouble will fail without calling the next result set
    expect( fastbit.result_set_next( resultHandle ) ).toEqual( 0 );
    expect( fastbit.result_set_getInt( resultHandle, 0 ) ).toEqual( 91 );
    expect( fastbit.result_set_getFloat( resultHandle, 1 ) ).toEqual( 91 );
    expect( fastbit.result_set_getInt( resultHandle, 2 ) ).toEqual( 8 );

    expect( fastbit.result_set_next_bundle( resultHandle ) ).toEqual( 0 );
    expect( fastbit.result_set_get_int( resultHandle, "c" ) ).toEqual( 7 );
    expect( fastbit.result_set_next_bundle( resultHandle ) ).toEqual( 0 );
    expect( fastbit.result_set_get_int( resultHandle, "c" ) ).toEqual( 6 );
    expect( fastbit.result_set_next_bundle( resultHandle ) ).toEqual( 0 );
    expect( fastbit.result_set_get_int( resultHandle, "c" ) ).toEqual( 5 );
    fastbit.destroy_query( queryHandle );
    fastbit.destroy_result_set( resultHandle );
  });

  it( 'should return a value for columns by name', function () {
    let queryHandle = uuid.v4();
    fastbit.build_query( queryHandle, 'a, b, c', "test/testdata", "a > 90 AND b < 95" );

    let resultHandle = uuid.v4();
    fastbit.build_result_set( queryHandle, resultHandle );
    expect( fastbit.result_set_next( resultHandle ) ).toEqual( 0 );
    expect( fastbit.result_set_get_double( resultHandle, 'a' ) ).toEqual( 91 );
    expect( fastbit.result_set_get_double( resultHandle, 'b' ) ).toEqual( 91 );
    expect( fastbit.result_set_get_double( resultHandle, 'c' ) ).toEqual( 8 );
    fastbit.destroy_query( queryHandle );
    fastbit.destroy_result_set( resultHandle );
  });

  it( 'should return a value for columns by alias', function () {
    let queryHandle = uuid.v4();
    fastbit.build_query( queryHandle, 'a as X, b as Y, c AS Z', "test/testdata", "a > 90 AND b < 95" );

    let resultHandle = uuid.v4();
    fastbit.build_result_set( queryHandle, resultHandle );
    expect( fastbit.result_set_next( resultHandle ) ).toEqual( 0 );
    expect( fastbit.result_set_get_double( resultHandle, 'X' ) ).toEqual( 91 );
    expect( fastbit.result_set_get_double( resultHandle, 'Y' ) ).toEqual( 91 );
    expect( fastbit.result_set_get_double( resultHandle, 'Z' ) ).toEqual( 8 );
    fastbit.destroy_query( queryHandle );
    fastbit.destroy_result_set( resultHandle );
  });
});

describe('insert operations', function () {
  // FIXME why does it fail to append when data directory is ./test/testdata
  // and works just fine when data directory is elsewhere?
  beforeEach( function () {
    fastbit.add_values('a', 'int',  testSetData[0], testSetData[0].length, 0);
    fastbit.add_values('b', 'float', testSetData[1], testSetData[1].length, 0);
    fastbit.add_values('c', 'short',  testSetData[2], testSetData[2].length, 0);
    expect( fastbit.flush_buffer("./testdata") ).toEqual(100);
    expect( fastbit.build_index( './test/testdata', 'a', 'none') ).toEqual(0);
    expect( fastbit.build_index( './test/testdata', 'b', 'none') ).toEqual(0);
    expect( fastbit.build_index( './test/testdata', 'c', 'none') ).toEqual(0);
  });
  afterEach( function (done) {
    fs.remove('./testdata', (err, data) => {
      if (err) return done(err);
      return done();
    })
  });

  it( 'should add new column to the end of existing rows in partition',
    function () {
    // Expect new column insertion to look like this:
    //       a    b  c  d
    //  0    #    #  #  -
    //  1    #    #  #  -
    //  #    .    .  .  -
    //  #    .    .  .  -
    //  #    .    .  .  -
    //  99   #    -  #  -
    // 100   100 1.0 1 '1'
    // 101   101 2.0 2 '2'
    // 102   102 3.0 3 '3'
    // 103   -    -  - '4'
    expect( fastbit.add_values("a", "int", [100,101,102], 3, 0) ).toEqual(0);
    expect( fastbit.add_values("b", "float", [1.0,2.0,3.0,4.0], 3, 0) ).toEqual(0);
    expect( fastbit.add_values("c", "short", [1,2,3,4], 3, 0) ).toEqual(0);
    expect( fastbit.add_values("d", "ubyte", ['1','2','3','4'], 4, 0) ).toEqual(0);
    // must flush in order for in-file DB to reflect in-memory table
    expect( fastbit.flush_buffer("./testdata") ).toEqual(4);

    let queryHandle = uuid.v4();
    fastbit.build_query( queryHandle, 'a, c', "./testdata", "a > 98" );
    let resultHandle = uuid.v4();
    fastbit.build_result_set( queryHandle, resultHandle );
    let rowsInA = fastbit.get_result_rows( queryHandle );
    let rowRes = fastbit.get_result_row_ids( queryHandle );
    expect( rowsInA ).toEqual( 4 );
    expect( rowRes.ret ).toEqual( 4 );
    expect( rowRes.ids ).toEqual( [99, 100, 101, 102] );
    fastbit.destroy_query( queryHandle );

    queryHandle = uuid.v4();
    fastbit.build_query( queryHandle, 'a, b, c, d', "./testdata", "a > 98" );
    rowsInA = fastbit.get_result_rows( queryHandle );
    rowRes = fastbit.get_result_row_ids( queryHandle );
    expect( rowsInA ).toEqual( 3 );
    expect( rowRes.ret ).toEqual( 3 );
    expect( rowRes.ids ).toEqual( [100, 101, 102] );
    fastbit.destroy_query( queryHandle );

    queryHandle = uuid.v4();
    fastbit.build_query( queryHandle, 'a', "./testdata", "a > 98" );
    rowsInA = fastbit.get_result_rows( queryHandle );
    rowRes = fastbit.get_result_row_ids( queryHandle );
    expect( rowsInA ).toEqual( 4 );
    expect( rowRes.ret ).toEqual( 4 );
    expect( rowRes.ids ).toEqual( [99, 100, 101, 102] );
    fastbit.destroy_query( queryHandle );

    queryHandle = uuid.v4();
    fastbit.build_query( queryHandle, 'd', "./testdata", "d = '4'" );
    rowsInA = fastbit.get_result_rows( queryHandle );
    rowRes = fastbit.get_result_row_ids( queryHandle );
    expect( rowsInA ).toEqual( 1 );
    expect( rowRes.ret ).toEqual( 1 );
    expect( rowRes.ids ).toEqual( [103] );
    fastbit.destroy_query( queryHandle );
  });
});

